from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.model_selection import train_test_split
from sklearn.metrics import roc_auc_score
import shap
import pandas as pd


class ExplanationAudit(BaseEstimator, ClassifierMixin):
    """
    Given a model, a dataset, and the protected attribute, we want to know if the model violates demographic parity or not and what are the features pushing for it.
    We do this by computing the shap values of the model, and then train a classifier to distinguish the protected attribute.

    Example
    -------
    >>> import pandas as pd
    >>> import numpy as np
    >>> from sklearn.linear_model import LogisticRegression
    >>> from xgboost import XGBRegressor
    >>> from fairtools.detector import ExplanationAudit
    >>> N = 5_000
    >>> x1 = np.random.normal(1, 1, size=N)
    >>> x2 = np.random.normal(1, 1, size=N)
    >>> x34 = np.random.multivariate_normal([1, 1], [[1, 0.5], [0.5, 1]], size=N)
    >>> x3 = x34[:, 0]
    >>> x4 = x34[:, 1]
    >>> # Binarize protected attribute
    >>> x4 = np.where(x4 > np.mean(x4), 1, 0)
    >>> X = pd.DataFrame([x1, x2, x3, x4]).T
    >>> X.columns = ["var%d" % (i + 1) for i in range(X.shape[1])]
    >>> y = (x1 + x2 + x3) / 3
    >>> y = 1 / (1 + np.exp(-y))
    >>> detector = ExplanationAudit(model=XGBRegressor(), gmodel=LogisticRegression())
    >>> detector.fit(X, y, Z="var4")
    >>> detector.get_auc_val()
    """

    def __init__(self, model, gmodel):
        self.model = model
        self.gmodel = gmodel
        self.explainer = None

        # Supported F Models
        self.supported_tree_models = [
            "XGBClassifier",
            "XGBRegressor",
            "GradientBoostingRegressor",
        ]
        self.supported_linear_models = [
            "LogisticRegression",
            "LinearRegression",
            "Ridge",
            "Lasso",
        ]
        self.supported_models = (
            self.supported_tree_models + self.supported_linear_models
        )
        # Supported detectors
        self.supported_linear_detectors = [
            "LogisticRegression",
        ]
        self.supported_tree_detectors = [
            "XGBClassifier",
            "GradientBoostingClassifier",
            "RandomForestClassifier",
        ]
        self.supported_detectors = (
            self.supported_linear_detectors + self.supported_tree_detectors
        )

        # Check if models are supported
        if self.get_model_type() not in self.supported_models:
            raise ValueError(
                "Model not supported. Supported models are: {} got {}".format(
                    self.supported_models, self.model.__class__.__name__
                )
            )
        if self.get_gmodel_type() not in self.supported_detectors:
            raise ValueError(
                "gmodel not supported. Supported models are: {} got {}".format(
                    self.supported_detectors, self.gmodel.__class__.__name__
                )
            )

    def get_gmodel_type(self):
        if self.gmodel.__class__.__name__ == "Pipeline":
            return self.gmodel.steps[-1][1].__class__.__name__
        else:
            return self.gmodel.__class__.__name__

    def get_model_type(self):
        if self.model.__class__.__name__ == "Pipeline":
            return self.model.steps[-1][1].__class__.__name__
        else:
            return self.model.__class__.__name__

    def get_split_data(self, X, y, Z, n1=0.6, n2=0.5):
        self.X_tr, X_val, self.y_tr, y_val = train_test_split(
            X, y, random_state=0, test_size=0.6, stratify=X[Z]
        )
        self.X_val, self.X_te, self.y_val, self.y_te = train_test_split(
            X_val, y_val, random_state=0, test_size=0.5, stratify=X_val[Z]
        )
        # Check number of classes present in label
        if len(set(self.y_tr)) <= 1:
            raise ValueError("Train set has only one class")
        if len(set(self.y_val)) <= 1:
            raise ValueError("Validation set has only one class")
        if len(set(self.y_te)) <= 1:
            raise ValueError("Test set has only one class")

        # Check number or protected groups
        if len(set(self.X_tr[Z])) <= 1:
            raise ValueError("Train set has only one protected group")
        if len(set(self.X_val[Z])) <= 1:
            raise ValueError("Validation set has only one protected group")
        if len(set(self.X_te[Z])) <= 1:
            raise ValueError("Test set has only one protected group")

        return self.X_tr, self.X_val, self.X_te, self.y_tr, self.y_val, self.y_te

    def fit(self, X, y, Z):

        # Check that X and y have correct shape
        check_X_y(X, y)
        self.Z = Z

        # Split data intro train, validation and test
        _ = self.get_split_data(X, y, Z)

        # Extract prottected att. and remove from data
        self.Z_tr = self.X_tr[self.Z]
        self.Z_val = self.X_val[self.Z]
        self.Z_te = self.X_te[self.Z]
        self.X_tr = self.X_tr.drop(self.Z, axis=1)
        self.X_val = self.X_val.drop(self.Z, axis=1)
        self.X_te = self.X_te.drop(self.Z, axis=1)

        # Fit model F
        self.fit_model(self.X_tr, self.y_tr)

        # Get explanations
        self.S_val = self.get_explanations(self.X_val)

        # Fit model G
        self.fit_audit_detector(self.S_val, self.Z_val)

        return self

    def predict(self, X):
        if self.Z in X.columns:
            X = X.drop(self.Z, axis=1)
        return self.gmodel.predict(self.get_explanations(X))

    def predict_proba(self, X):
        if self.Z in X.columns:
            X = X.drop(self.Z, axis=1)
        return self.gmodel.predict_proba(self.get_explanations(X))

    def explanation_predict(self, X):
        return self.gmodel.predict(X)

    def explanation_predict_proba(self, X):
        return self.gmodel.predict_proba(X)

    def fit_model(self, X, y):
        self.model.fit(X, y)

    def fit_audit_detector(self, X, y):
        self.gmodel.fit(X, y)

    def get_explanations(self, X):
        # Determine the type of SHAP explainer to use
        if self.get_model_type() in self.supported_tree_models:
            self.explainer = shap.Explainer(self.model)
        elif self.get_model_type() in self.supported_linear_models:
            self.explainer = shap.LinearExplainer(
                self.model, X, feature_dependence="correlation_dependent"
            )
        else:
            raise ValueError(
                "Model not supported. Supported models are: {}, got {}".format(
                    self.supported_models, self.model.__class__.__name__
                )
            )

        shap_values = self.explainer(X)
        # Name columns
        if isinstance(X, pd.DataFrame):
            columns_name = X.columns
        else:
            columns_name = ["Shap%d" % (i + 1) for i in range(X.shape[1])]

        exp = pd.DataFrame(
            data=shap_values.values,
            columns=columns_name,
        )
        return exp

    def get_auc_f_val(self):
        """
        TODO Case of F being a classifier
        """
        return roc_auc_score(self.y_val, self.model.predict(self.X_val))

    def get_auc_val(self):
        """
        Returns the AUC of the validation set

        """
        return roc_auc_score(self.Z_te, self.predict_proba(self.X_te)[:, 1])

    def get_coefs(self):
        if self.gmodel.__class__.__name__ == "Pipeline":
            if (
                self.gmodel.steps[-1][1].__class__.__name__
                in self.supported_linear_models
            ):
                return self.gmodel.steps[-1][1].coef_
            else:
                raise ValueError(
                    "Pipeline model not supported. Supported models are: {}, got {}".format(
                        self.supported_linear_models,
                        self.gmodel.steps[-1][1].__class__.__name__,
                    )
                )
        else:
            return self.get_linear_coefs()

    def get_linear_coefs(self):
        if self.gmodel.__class__.__name__ in self.supported_linear_models:
            return self.gmodel.coef_[0]
        else:
            raise ValueError(
                "Detector model not supported. Supported models ar linear: {}, got {}".format(
                    self.supported_linear_detector, self.model.__class__.__name__
                )
            )



from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.model_selection import train_test_split
from sklearn.metrics import roc_auc_score
import pandas as pd
import shap


class ExplanationShiftDetector(BaseEstimator, ClassifierMixin):
    """
    Given a model, and two datasets (source,test), we want to know if the behaviour of the model is different bt train and test.
    We can do this by computing the shap values of the model on the two datasets, and then train a classifier to distinguish between the two datasets.

    Example
    -------
    >>> import pandas as pd
    >>> from sklearn.model_selection import train_test_split
    >>> from sklearn.datasets import make_blobs
    >>> from tools.xaiUtils import ExplanationShiftDetector
    >>> from xgboost import XGBRegressor
    >>> from sklearn.linear_model import LogisticRegression

    >>> X, y = make_blobs(n_samples=2000, centers=2, n_features=5, random_state=0)
    >>> X_tr, X_te, y_tr, y_te = train_test_split(X, y, test_size=0.5, random_state=0)
    >>> X_ood,y_ood = make_blobs(n_samples=1000, centers=1, n_features=5, random_state=0)

    >>> detector = ExplanationShiftDetector(model=XGBRegressor(),gmodel=LogisticRegression())
    >>> detector.fit(X_tr, y_tr, X_ood)
    >>> detector.get_auc_val()
    # 0.76
    >>> detector.fit(X_tr, y_tr, X_te)
    >>> detector.get_auc_val()
    # 0.5
    """

    def __init__(
        self,
        model,
        gmodel,
        space: str = "explanation",
        algorithm: str = "auto",
        masker: bool = False,
    ):
        """
        Parameters
        ----------
        model : sklearn model
            Model to be used to compute the shap values.
        gmodel : sklearn model
            Model to be used to distinguish between the two datasets.
        space : str, optional
            Space in which the gmodel is learned. Can be 'explanation' or 'input' or 'predictions'. Default is 'explanation'.

        algorithm : "auto", "permutation", "partition", "tree", or "linear"
                The algorithm used to estimate the Shapley values. There are many different algorithms that
                can be used to estimate the Shapley values (and the related value for constrained games), each
                of these algorithms have various tradeoffs and are preferrable in different situations. By
                default the "auto" options attempts to make the best choice given the passed model and masker,
                but this choice can always be overriden by passing the name of a specific algorithm. The type of
                algorithm used will determine what type of subclass object is returned by this constructor, and
                you can also build those subclasses directly if you prefer or need more fine grained control over
                their options.

        masker : bool,
                The masker object is used to define the background distribution over which the Shapley values
                are estimated. Is a boolean that indicates if the masker should be used or not. If True, the masker is used.
                If False, the masker is not used. The background distribution is the same distribution as we are calculating the Shapley values.
                TODO Decide which masker distribution is better to use, options are: train data, hold out data, ood data
        """

        self.model = model
        self.gmodel = gmodel
        self.explainer = None
        self.space = space
        self.algorithm = algorithm
        self.masker = masker

        # Check if space is supported
        if self.space not in ["explanation", "input", "prediction"]:
            raise ValueError(
                "space not supported. Supported spaces are: {} got {}".format(
                    ["explanation", "input", "prediction"], self.space
                )
            )

    def get_gmodel_type(self):
        """
        Returns the type of the gmodel
        """
        if self.gmodel.__class__.__name__ == "Pipeline":
            return self.gmodel.steps[-1][1].__class__.__name__
        else:
            return self.gmodel.__class__.__name__

    def get_model_type(self):
        if self.model.__class__.__name__ == "Pipeline":
            return self.model.steps[-1][1].__class__.__name__
        else:
            return self.model.__class__.__name__

    def fit(self, X_source, y_source, X_ood):
        """
        1. Fits the model F to the data by splitting the data into two equal parts.
        2. Get the explanations of the model F on the validation set and OOD
        3. Fit the inspector to the data to the data.

        """

        # Check that X and y have correct shape
        check_X_y(X_source, y_source)
        self.X_ood = X_ood

        self.X_tr, self.X_val, self.y_tr, self.y_val = train_test_split(
            X_source, y_source, random_state=0, test_size=0.5
        )

        # Fit model F
        self.fit_model(self.X_tr, self.y_tr)

        # Get explanations
        self.S_val = self.get_explanations(self.X_val)
        self.S_ood = self.get_explanations(self.X_ood)

        # Create dataset for  explanation shift detector
        self.S_val["label"] = False
        self.S_ood["label"] = True

        self.S = pd.concat([self.S_val, self.S_ood])

        (
            self.X_shap_tr,
            self.X_shap_te,
            self.y_shap_tr,
            self.y_shap_te,
        ) = train_test_split(
            self.S.drop(columns="label"), self.S["label"], random_state=0, test_size=0.5
        )
        self.fit_explanation_shift(self.X_shap_tr, self.y_shap_tr)

        return self

    def predict(self, X):
        return self.gmodel.predict(self.get_explanations(X))

    def predict_proba(self, X):
        return self.gmodel.predict_proba(self.get_explanations(X))

    def explanation_predict(self, X):
        return self.gmodel.predict(X)

    def explanation_predict_proba(self, X):
        return self.gmodel.predict_proba(X)

    def fit_model(self, X, y):
        self.model.fit(X, y)

    def fit_explanation_shift(self, X, y):
        self.gmodel.fit(X, y)

    def get_explanations(self, X, data_masker=None):
        if data_masker == None:
            data_masker = self.X_tr
        else:
            data_masker = data_masker

        if self.space == "explanation":
            if self.masker:
                self.explainer = shap.Explainer(
                    self.model, algorithm=self.algorithm, masker=data_masker
                )
            else:
                self.explainer = shap.Explainer(self.model, algorithm=self.algorithm)

            shap_values = self.explainer(X)
            # Name columns
            if isinstance(X, pd.DataFrame):
                columns_name = X.columns
            else:
                columns_name = ["Shap%d" % (i + 1) for i in range(X.shape[1])]

            exp = pd.DataFrame(
                data=shap_values.values,
                columns=columns_name,
            )
        if self.space == "input":
            shap_values = X
            # Name columns
            if isinstance(X, pd.DataFrame):
                exp = X
            else:
                columns_name = ["Shap%d" % (i + 1) for i in range(X.shape[1])]

                exp = pd.DataFrame(
                    data=shap_values,
                    columns=columns_name,
                )
        if self.space == "prediction":
            try:
                shap_values = self.model.predict_proba(X)[:, 1]
            except:
                shap_values = self.model.predict(X)

            # Name columns
            exp = pd.DataFrame(
                data=shap_values,
                columns=["preds"],
            )

        return exp

    def get_auc_val(self):
        """
        Returns the AUC of the explanation shift detector on the validation set of the explanation space
        Example
        -------
        from sklearn.model_selection import train_test_split
        from sklearn.datasets import make_blobs
        from tools.xaiUtils import ExplanationShiftDetector
        from xgboost import XGBRegressor
        from sklearn.linear_model import LogisticRegression

        # Create data
        X, y = make_blobs(n_samples=2000, centers=2, n_features=5, random_state=0)
        X_tr, X_te, y_tr, y_te = train_test_split(X, y, test_size=0.5, random_state=0)
        X_ood,y_ood = make_blobs(n_samples=1000, centers=1, n_features=5, random_state=0)

        detector = ExplanationShiftDetector(model=XGBRegressor(),gmodel=LogisticRegression())
        detector.fit(X_tr, y_tr, X_ood)
        detector.get_auc_val()
        # 0.76

        """
        return roc_auc_score(
            self.y_shap_te, self.explanation_predict_proba(self.X_shap_te)[:, 1]
        )

    def get_coefs(self):
        if self.gmodel.__class__.__name__ == "Pipeline":
            if "sklearn.linear_model" in self.gmodel.steps[-1][1].__module__:
                return self.gmodel.steps[-1][1].coef_
            else:
                raise ValueError(
                    "Coefficients can not be calculated. Supported models are linear: sklearn.linear_model, got {}".format(
                        self.gmodel.steps[-1][1].__module__
                    )
                )
        else:
            return self.get_linear_coefs()

    def get_linear_coefs(self):
        if "sklearn.linear_model" in self.gmodel.__module__:
            return self.gmodel.coef_
        else:
            raise ValueError(
                "Coefficients can not be calculated. Supported models are linear: sklearn.linear_model, got {}".format(
                    self.gmodel.steps[-1][1].__module__
                )
            )

    def explain_detector(self):
        if self.space == "prediction":
            return
        exp = shap.Explainer(self.model)

        shap_values = exp.shap_values(self.S_ood.drop(columns="label"))
        shap.summary_plot(
            shap_values, self.S_ood.drop(columns="label"), plot_type="bar", show=False
        )



from nobias import ExplanationShiftDetector


from xgboost import XGBClassifier, XGBRegressor
from sklearn.linear_model import LogisticRegression, LinearRegression
from sklearn.model_selection import train_test_split
from sklearn.datasets import make_blobs
import pandas as pd
import numpy as np
import pdb


X, y = make_blobs(n_samples=2000, centers=2, n_features=5, random_state=0)
X_tr, X_te, y_tr, y_te = train_test_split(X, y, test_size=0.5, random_state=0)
X_ood, y_ood = make_blobs(n_samples=1000, centers=1, n_features=5, random_state=0)


def test_return_shapDF():
    """
    If X is dataframe, return shap values as dataframe.
    """
    XX = pd.DataFrame(X, columns=["a", "b", "c", "d", "e"])
    XX_ood = pd.DataFrame(X_ood, columns=["a", "b", "c", "d", "e"])

    esd = ExplanationShiftDetector(
        model=LogisticRegression(), gmodel=LogisticRegression(), masker=True
    )
    esd.fit(XX, y, XX_ood)
    ex = esd.get_explanations(XX)
    assert all([a == b for a, b in zip(ex.columns, XX.columns)])


def test_doc_examples():
    """
    Check that doc examples work.
    """

    detector = ExplanationShiftDetector(
        model=XGBRegressor(random_state=0), gmodel=LogisticRegression()
    )
    # On OOD
    detector.fit(X_source=X_tr, y_source=y_tr, X_ood=X_ood)
    assert np.round(detector.get_auc_val(), decimals=2) == 0.77
    # On test
    detector.fit(X_source=X_tr, y_source=y_tr, X_ood=X_te)
    assert np.round(detector.get_auc_val(), decimals=2) == 0.53


def test_no_nan():
    """
    Check that no NaNs are present in the shap values.
    """
    esd = ExplanationShiftDetector(model=XGBClassifier(), gmodel=XGBClassifier())
    esd.fit(X, y, X_ood)
    ex = esd.get_explanations(X)
    assert not np.any(np.isnan(ex))


def test_get_coefs_linear():
    """
    Check that the coefficients are returned correctly for the linear regression.
    """
    esd = ExplanationShiftDetector(
        model=LinearRegression(), gmodel=LogisticRegression(), masker=True
    )
    esd.fit(X, y, X_ood)
    coefs = esd.get_linear_coefs()
    # Assert shape
    assert coefs.shape[1] == X.shape[1]
    # Assert that there is non NaNs
    assert not np.any(np.isnan(coefs))
    # Check when we call the full methods
    coefs = esd.get_coefs()
    # Assert shape
    assert coefs.shape[1] == X.shape[1]
    # Assert that there is non NaNs
    assert not np.any(np.isnan(coefs))


def test_get_coefs_pipeline():
    """
    Check that the coefficients are returned correctly for the linear regression pipeline.
    TODO : add a test for the case of a pipeline for F.
    """
    from sklearn.pipeline import Pipeline
    from sklearn.preprocessing import StandardScaler

    esd = ExplanationShiftDetector(
        model=LinearRegression(),
        gmodel=Pipeline([("scaler", StandardScaler()), ("lr", LogisticRegression())]),
        masker=True,
    )
    esd.fit(X_tr, y_tr, X_ood)
    coefs = esd.get_coefs()
    # Assert shape
    assert coefs.shape[1] == X.shape[1]
    # Assert that there is non NaNs
    assert not np.any(np.isnan(coefs))


def test_get_model_types():
    """
    Check that the model types are returned correctly.
    """
    from sklearn.pipeline import Pipeline
    from sklearn.preprocessing import StandardScaler

    esd = ExplanationShiftDetector(
        model=LinearRegression(), gmodel=LogisticRegression(), masker=X
    )
    assert esd.get_gmodel_type(), esd.get_model_type() == ("linear", "linear")
    # Case of pipeline
    esd = ExplanationShiftDetector(
        model=Pipeline([("scaler", StandardScaler()), ("lr", LinearRegression())]),
        gmodel=Pipeline([("scaler", StandardScaler()), ("lr", LogisticRegression())]),
    )
    assert esd.get_gmodel_type(), esd.get_model_type() == ("linear", "linear")


def test_spaces():
    """
    Check that the spaces are returned correctly.
    """

    esd = ExplanationShiftDetector(
        model=LinearRegression(),
        gmodel=LogisticRegression(),
        space="input",
        masker=True,
    )
    esd.fit(X_tr, y_tr, X_ood)
    # Check if returns input space
    # assert esd.get_explanations(X) == X
    np.testing.assert_array_equal(esd.get_explanations(X), X)
    # Check if returns output space
    esd = ExplanationShiftDetector(
        model=LogisticRegression(),
        gmodel=LogisticRegression(),
        space="prediction",
        masker=True,
    )
    esd.fit(X, y, X_ood)
    np.testing.assert_array_equal(
        esd.get_explanations(X).shape,
        pd.DataFrame(data=esd.model.predict(X), columns=["preds"]).shape,
    )

    # Check if returns exp space
    esd = ExplanationShiftDetector(
        model=LinearRegression(),
        gmodel=LogisticRegression(),
        space="explanation",
        masker=True,
    )
    esd.fit(X, y, X_ood)

    np.testing.assert_array_equal(esd.get_explanations(X).shape, X.shape)


def test_tree_shap():
    """
    Check that the shap values are returned correctly for the tree models.
    """
    esd = ExplanationShiftDetector(
        model=XGBRegressor(), gmodel=LogisticRegression(), masker=True
    )
    esd.fit(X, y, X_ood)
    shap_values = esd.get_explanations(X)
    # Assert shape
    assert shap_values.shape[1] == X.shape[1]
    # Assert that there is non NaNs
    assert not np.any(np.isnan(shap_values))

    esd = ExplanationShiftDetector(model=XGBRegressor(), gmodel=LogisticRegression())
    esd.fit(X, y, X_ood)
    shap_values2 = esd.get_explanations(X)
    # Assert shape
    assert shap_values2.shape[1] == X.shape[1]
    # Assert that there is non NaNs
    assert not np.any(np.isnan(shap_values2))
    # Assert that the shap values are different depending on the masker
    assert shap_values2.sum(axis=1).sum(axis=0) != shap_values.sum(axis=1).sum(axis=0)


def test_explain_detector():
    for space in ["input", "prediction", "explanation"]:
        esd = ExplanationShiftDetector(
            model=XGBClassifier(), gmodel=XGBClassifier(), space=space
        )
        esd.fit(X, y, X_ood)
        esd.explain_detector()



from nobias import ExplanationAudit

from sklearn.ensemble import GradientBoostingClassifier, GradientBoostingRegressor
from sklearn.linear_model import LogisticRegression, LinearRegression
from sklearn.model_selection import train_test_split
from sklearn.datasets import make_blobs
import pandas as pd
import numpy as np


X, y = make_blobs(n_samples=1000, centers=2, n_features=5, random_state=0)
X = pd.DataFrame(X, columns=["a", "b", "c", "d", "e"])
# Protected att
X["a"] = np.where(X["a"] > X["a"].mean(), 1, 0)
X_tr, X_te, y_tr, y_te = train_test_split(X, y, test_size=0.5, random_state=0)


def test_get_splits():
    detector = ExplanationAudit(
        model=GradientBoostingRegressor(), gmodel=LogisticRegression()
    )
    N = 1000
    X, y = make_blobs(n_samples=N, centers=2, n_features=5, random_state=0)
    X = pd.DataFrame(X, columns=["a", "b", "c", "d", "e"])
    # Binarize
    X["a"] = np.where(X["a"] > X["a"].mean(), 1, 0)

    detector.get_split_data(X, y, Z="a", n1=0.6, n2=0.5)
    assert detector.X_tr.shape == (N * 0.4, 5)
    assert detector.X_val.shape == (N * 0.6 * 0.5, 5)
    assert detector.X_te.shape == (N * 0.6 * 0.5, 5)

    assert len(set(detector.y_tr)) > 1
    assert len(set(detector.y_val)) > 1
    assert len(set(detector.y_te)) > 1


def test_return_shapDF():
    """
    If X is dataframe, return shap values as dataframe.
    """
    XX = pd.DataFrame(X, columns=["a", "b", "c", "d", "e"])

    esd = ExplanationAudit(model=LinearRegression(), gmodel=LogisticRegression())
    esd.fit_model(XX, y)
    ex = esd.get_explanations(XX)
    assert all([a == b for a, b in zip(ex.columns, XX.columns)])


def test_supported_models():
    """
    Check that models are supported.
    """
    for model in [GradientBoostingRegressor(), LogisticRegression()]:
        for gmodel in [GradientBoostingClassifier(), LogisticRegression()]:
            assert (
                type(ExplanationAudit(model=model, gmodel=gmodel)) is ExplanationAudit
            )


def test_not_supported_models():
    """
    Check that models are not supported.
    """

    from sklearn.neural_network import MLPClassifier
    import pytest

    with pytest.raises(ValueError):
        ExplanationAudit(model=MLPClassifier(), gmodel=LogisticRegression())
    with pytest.raises(ValueError):
        ExplanationAudit(model=LinearRegression(), gmodel=MLPClassifier())


def test_get_model_types():
    """
    Check that the model types are returned correctly.
    """
    from sklearn.pipeline import Pipeline
    from sklearn.preprocessing import StandardScaler

    esd = ExplanationAudit(model=LinearRegression(), gmodel=LogisticRegression())
    assert esd.get_gmodel_type(), esd.get_model_type() == ("linear", "linear")
    # Case of pipeline
    esd = ExplanationAudit(
        model=Pipeline([("scaler", StandardScaler()), ("lr", LinearRegression())]),
        gmodel=Pipeline([("scaler", StandardScaler()), ("lr", LogisticRegression())]),
    )
    assert esd.get_gmodel_type(), esd.get_model_type() == ("linear", "linear")


def test_no_nan():
    """
    Check that no NaNs are present in the shap values.
    """
    esd = ExplanationAudit(model=LinearRegression(), gmodel=LogisticRegression())
    esd.fit_model(X, y)
    ex = esd.get_explanations(X)
    assert not np.any(np.isnan(ex))


def test_get_coefs_linear():
    """
    Check that the coefficients are returned correctly for the linear regression.
    """
    esd = ExplanationAudit(model=LinearRegression(), gmodel=LogisticRegression())
    esd.fit(X, y, Z="a")
    coefs = esd.get_linear_coefs()
    # Assert shape -1 from protected attribute
    assert len(coefs) == X.shape[1] - 1
    # Assert that there is non NaNs
    assert not np.any(np.isnan(coefs))
    # Check when we call the full methods
    coefs = esd.get_coefs()
    # Assert shape -1 from protected attribute
    assert len(coefs) == X.shape[1] - 1
    # Assert that there is non NaNs
    assert not np.any(np.isnan(coefs))


def test_get_coefs_pipeline():
    """
    Check that the coefficients are returned correctly for the linear regression pipeline.
    TODO : add a test for the case of a pipeline for F.
    """
    from sklearn.pipeline import Pipeline
    from sklearn.preprocessing import StandardScaler

    esd = ExplanationAudit(
        model=LinearRegression(),
        gmodel=Pipeline([("scaler", StandardScaler()), ("lr", LogisticRegression())]),
    )
    esd.fit(X, y, Z="a")
    coefs = esd.get_coefs()
    # Assert shape -1 from protected attribute
    assert coefs.shape[1] == X.shape[1] - 1
    # Assert that there is non NaNs
    assert not np.any(np.isnan(coefs))


def test_predict_drop_prottected():
    """
    Check that the protected attribute is dropped when predicting.
    """
    audit = ExplanationAudit(
        model=GradientBoostingRegressor(), gmodel=LogisticRegression()
    )

    audit.fit(X, y, Z="a")
    assert np.isnan(audit.predict(X)).sum() == 0
    assert np.isnan(audit.predict_proba(X)).sum() == 0


'''
def test_doc_examples():
    """
    Check that doc examples work.
    WARNING: this test takes a long time to run. (1-5mins)
    """
    from folktables import ACSDataSource, ACSIncome

    data_source = ACSDataSource(survey_year="2014", horizon="1-Year", survey="person")
    try:
        acs_data = data_source.get_data(states=["CA"], download=False)
    except:
        acs_data = data_source.get_data(states=["CA"], download=True)

    X_, y_, _ = ACSIncome.df_to_numpy(acs_data)
    X_ = pd.DataFrame(X_, columns=ACSIncome.features)

    # White vs ALL
    X_["RAC1P"] = np.where(X_["RAC1P"] == 1, 1, 0)

    detector = ExplanationAudit(
        model=GradientBoostingRegressor(random_state=0), gmodel=LogisticRegression()
    )

    detector.fit(X_, y_, Z="RAC1P")
    # Check that the model prediction works
    assert np.round(detector.get_auc_val(), decimals=1) == 0.7
    # Check that the coefficients are returned correctly
    coefs = detector.get_coefs()
    assert len(coefs) == X_.shape[1] - 1  # -1 for the protected attribute
    assert np.isnan(coefs).sum() == 0
    # Hard coded results.
    # assert np.round(coefs,decimals=0) ==[ 1., -2., -1.,  1., -0., 17.,  2., -0.,  1.]
'''
